@@ -15,27 +15,30 @@
 #
 # You should have received a copy of the GNU Lesser General Public License along
 # with astroid. If not, see <http://www.gnu.org/licenses/>.
-from os.path import join, abspath, dirname
 import sys
 import unittest
+import textwrap
 
-from astroid import ResolveError, MANAGER, Instance, nodes, YES, InferenceError
+from astroid import MANAGER, Instance, nodes
 from astroid.builder import AstroidBuilder
 from astroid.raw_building import build_module
 from astroid.manager import AstroidManager
 from astroid.test_utils import require_version
 from astroid.tests import resources
 
-class NonRegressionTests(unittest.TestCase):
+class NonRegressionTests(resources.AstroidCacheSetupMixin,
+                         unittest.TestCase):
 
     def setUp(self):
         sys.path.insert(0, resources.find('data'))
+        MANAGER.always_load_extensions = True
 
     def tearDown(self):
         # Since we may have created a brainless manager, leading
         # to a new cache builtin module and proxy classes in the constants,
         # clear out the global manager cache.
-        MANAGER.clear_cache()
+        MANAGER.clear_cache(self._builtins)
+        MANAGER.always_load_extensions = False
         sys.path.pop(0)
         sys.path_importer_cache.pop(resources.find('data'), None)
 
@@ -84,7 +87,7 @@ class NonRegressionTests(unittest.TestCase):
 
     def test_new_style_class_detection(self):
         try:
-            import pygtk
+            import pygtk # pylint: disable=unused-variable
         except ImportError:
             self.skipTest('test skipped: pygtk is not available')
         # XXX may fail on some pygtk version, because objects in
@@ -105,7 +108,7 @@ class A(gobject.GObject):
 
     def test_pylint_config_attr(self):
         try:
-            from pylint import lint
+            from pylint import lint # pylint: disable=unused-variable
         except ImportError:
             self.skipTest('pylint not available')
         mod = MANAGER.ast_from_module_name('pylint.lint')
@@ -126,7 +129,7 @@ class A(gobject.GObject):
         #a crash occured somewhere in the past, and an
         # InferenceError instead of a crash was better, but now we even infer!
         try:
-            import numpy
+            import numpy # pylint: disable=unused-variable
         except ImportError:
             self.skipTest('test skipped: numpy is not available')
         builder = AstroidBuilder()
@@ -139,7 +142,6 @@ multiply(1, 2, 3)
         callfunc = astroid.body[1].value.func
         infered = callfunc.infered()
         self.assertEqual(len(infered), 1)
-        self.assertIs(infered[0], YES)
 
     @require_version('3.0')
     def test_nameconstant(self):
@@ -218,6 +220,28 @@ def test():
         base = next(result._proxied.bases[0].infer())
         self.assertEqual(base.name, 'int')
 
+    def test_ancestors_patching_class_recursion(self):
+        node = AstroidBuilder().string_build(textwrap.dedent("""
+        import string
+        Template = string.Template
+
+        class A(Template):
+            pass
+
+        class B(A):
+            pass
+
+        def test(x=False):
+            if x:
+                string.Template = A
+            else:
+                string.Template = B
+        """))
+        klass = node['A']
+        ancestors = list(klass.ancestors())
+        self.assertEqual(ancestors[0].qname(), 'string.Template')
+
+
 class Whatever(object):
     a = property(lambda x: x, lambda x: x)
 
